from typing import Iterable
from functools import cached_property
from dataclasses import dataclass

from kirin import ir
from kirin.print import Printer, Printable
from kirin.worklist import WorkList


@dataclass
class CFG(Printable):
    """Control Flow Graph of a given IR statement.

    This class implements the [`kirin.graph.Graph`][kirin.graph.Graph] protocol.

    !!! note "Pretty Printing"
        This object is pretty printable via
        [`.print()`][kirin.print.printable.Printable.print] method.
    """

    parent: ir.Region
    """Parent IR statement.
    """
    entry: ir.Block | None = None
    """Entry block of the CFG.
    """

    def __post_init__(self):
        if self.parent.blocks.isempty():
            self.entry = None
        else:
            self.entry = self.parent.blocks[0]

    @cached_property
    def predecessors(self):
        """CFG data, mapping a block to its predecessors."""
        graph: dict[ir.Block, set[ir.Block]] = {}
        for block, neighbors in self.successors.items():
            for neighbor in neighbors:
                graph.setdefault(neighbor, set()).add(block)
        return graph

    @cached_property
    def successors(self):
        """CFG data, mapping a block to its neighbors."""
        graph: dict[ir.Block, set[ir.Block]] = {}
        visited: set[ir.Block] = set()
        worklist: WorkList[ir.Block] = WorkList()
        if self.parent.blocks.isempty():
            return graph

        block = self.entry
        while block is not None:
            neighbors = graph.setdefault(block, set())
            if block.last_stmt is not None:
                neighbors.update(block.last_stmt.successors)
                worklist.extend(block.last_stmt.successors)
            visited.add(block)

            block = worklist.pop()
            while block is not None and block in visited:
                block = worklist.pop()
        return graph

    # graph interface
    def get_neighbors(self, node: ir.Block) -> Iterable[ir.Block]:
        return self.successors[node]

    def get_edges(self) -> Iterable[tuple[ir.Block, ir.Block]]:
        for block, neighbors in self.successors.items():
            for neighbor in neighbors:
                yield block, neighbor

    def get_nodes(self) -> Iterable[ir.Block]:
        return self.successors.keys()

    # printable interface
    def print_impl(self, printer: Printer) -> None:
        # NOTE: this make sure we use the same name
        # as the printing of CFG parent.
        with printer.string_io():
            self.parent.print(printer)

        printer.plain_print("Successors:")
        printer.print_newline()
        for block, neighbors in self.successors.items():
            printer.plain_print(f"{printer.state.block_id[block]} -> ", end="")
            printer.print_seq(
                neighbors,
                delim=", ",
                prefix="[",
                suffix="]",
                emit=lambda block: printer.plain_print(printer.state.block_id[block]),
            )
            printer.print_newline()

        if self.predecessors:
            printer.print_newline()
            printer.plain_print("Predecessors:")
            printer.print_newline()
            for block, neighbors in self.predecessors.items():
                printer.plain_print(f"{printer.state.block_id[block]} <- ", end="")
                printer.print_seq(
                    neighbors,
                    delim=", ",
                    prefix="[",
                    suffix="]",
                    emit=lambda block: printer.plain_print(
                        printer.state.block_id[block]
                    ),
                )
                printer.print_newline()



"""This module provides a function to create a function dynamically.

Copied from `dataclasses._create_fn` in Python 3.10.13.
"""

from dataclasses import MISSING


def create_fn(name, args, body, *, globals=None, locals=None, return_type=MISSING):
    # Note that we may mutate locals. Callers beware!
    # The only callers are internal to this module, so no
    # worries about external callers.
    if locals is None:
        locals = {}
    return_annotation = ""
    if return_type is not MISSING:
        locals["_return_type"] = return_type
        return_annotation = "->_return_type"
    args = ",".join(args)
    body = "\n".join(f"        {b}" for b in body)

    # Compute the text of the entire function.
    txt = f"    def {name}({args}){return_annotation}:\n{body}"

    local_vars = ", ".join(locals.keys())
    txt = f"def __create_fn__({local_vars}):\n{txt}\n    return {name}"
    ns = {}
    exec(txt, globals, ns)
    return ns["__create_fn__"](**locals)



from kirin import interp
from kirin.analysis import const

from . import stmts
from ._dialect import dialect


@dialect.register(key="constprop")
class ConstProp(interp.MethodTable):

    @interp.impl(stmts.Not)
    def not_(
        self, _: const.Propagate, frame: const.Frame, stmt: stmts.Not
    ) -> interp.StatementResult[const.Result]:
        hint = frame.get(stmt.value)
        if isinstance(hint, (const.PartialTuple, const.Value)):
            ret = const.Value(not hint.data)
        else:
            ret = const.Unknown()
        return (ret,)



from abc import ABC, abstractmethod
from typing import Any, Generic, TypeVar, Iterable
from dataclasses import field, dataclass

from typing_extensions import Self

from kirin.ir import Block, SSAValue, Statement

from .exceptions import InterpreterError

ValueType = TypeVar("ValueType")


@dataclass
class FrameABC(ABC, Generic[ValueType]):
    """Abstract base class for the IR interpreter frame.

    While the IR is in SSA form which does not have the need
    of scoping, the frame is still useful to keep track of the
    current statement being interpreted and the call stack. As
    well as various other interpreter state based on the specific
    interpreter implementation.

    This base class provides the minimum interface for the
    interpreter frame.
    """

    code: Statement
    """statement whose region is being interpreted, e.g a function.
    """
    parent: Self | None = field(default=None, kw_only=True, compare=True, repr=False)
    """Parent frame.
    """
    has_parent_access: bool = field(default=False, kw_only=True, compare=True)
    """If we have access to the entries of the parent frame."""

    lineno_offset: int = field(default=0, kw_only=True, compare=True)

    current_stmt: Statement | None = field(
        default=None, init=False, compare=False, repr=False
    )
    """Current statement being interpreted."""
    current_block: Block | None = field(
        default=None, init=False, compare=False, repr=False
    )
    """Current block being interpreted."""

    @abstractmethod
    def get(self, key: SSAValue) -> ValueType:
        """Get the value for the given [`SSAValue`][kirin.ir.SSAValue] key.
        See also [`get_values`][kirin.interp.frame.Frame.get_values].

        Args:
            key(SSAValue): The key to get the value for.

        Returns:
            ValueType: The value.
        """
        ...

    @abstractmethod
    def set(self, key: SSAValue, value: ValueType) -> None:
        """Set the value for the given [`SSAValue`][kirin.ir.SSAValue] key.
        See also [`set_values`][kirin.interp.frame.Frame.set_values].

        Args:
            key(SSAValue): The key to set the value for.
            value(ValueType): The value.
        """
        ...

    def get_values(self, keys: Iterable[SSAValue]) -> tuple[ValueType, ...]:
        """Get the values of the given [`SSAValue`][kirin.ir.SSAValue] keys.
        See also [`get`][kirin.interp.frame.Frame.get].

        Args:
            keys(Iterable[SSAValue]): The keys to get the values for.

        Returns:
            tuple[ValueType, ...]: The values.
        """
        return tuple(self.get(key) for key in keys)

    def set_values(self, keys: Iterable[SSAValue], values: Iterable[ValueType]) -> None:
        """Set the values of the given [`SSAValue`][kirin.ir.SSAValue] keys.
        This is a convenience method to set multiple values at once.

        Args:
            keys(Iterable[SSAValue]): The keys to set the values for.
            values(Iterable[ValueType]): The values.
        """
        for key, value in zip(keys, values):
            self.set(key, value)


@dataclass
class Frame(FrameABC[ValueType]):
    """Interpreter frame."""

    globals: dict[str, Any] = field(default_factory=dict, kw_only=True)
    """Global variables this frame has access to.
    """

    # NOTE: we are sharing the same frame within blocks
    # this is because we are validating e.g SSA value pointing
    # to other blocks separately. This avoids the need
    # to have a separate frame for each block.
    entries: dict[SSAValue, ValueType] = field(default_factory=dict, kw_only=True)
    """SSA values and their corresponding values.
    """

    def get(self, key: SSAValue) -> ValueType:
        """Get the value for the given [`SSAValue`][kirin.ir.SSAValue].

        Args:
            key(SSAValue): The key to get the value for.

        Returns:
            ValueType: The value.

        Raises:
            InterpreterError: If the value is not found. This will be catched by the interpreter.
        """
        err = InterpreterError(f"SSAValue {key} not found")
        value = self.entries.get(key, err)
        if isinstance(value, InterpreterError):
            if self.has_parent_access and self.parent:
                return self.parent.get(key)
            else:
                raise err
        else:
            return value

    ExpectedType = TypeVar("ExpectedType")

    def get_casted(self, key: SSAValue, type_: type[ExpectedType]) -> ExpectedType:
        """Same as [`get`][kirin.interp.frame.Frame.get] except it
        forces the linter to think the value is of the expected type.

        Args:
            key(SSAValue): The key to get the value for.
            type_(type): The expected type.

        Returns:
            ExpectedType: The value.
        """
        return self.get(key)  # type: ignore

    def get_typed(self, key: SSAValue, type_: type[ExpectedType]) -> ExpectedType:
        """Similar to [`get`][kirin.interp.frame.Frame.get] but also checks the type.

        Args:
            key(SSAValue): The key to get the value for.
            type_(type): The expected type.

        Returns:
            ExpectedType: The value.

        Raises:
            InterpreterError: If the value is not of the expected type.
        """
        value = self.get(key)
        if not isinstance(value, type_):
            raise InterpreterError(f"expected {type_}, got {type(value)}")
        return value

    def set(self, key: SSAValue, value: ValueType) -> None:
        self.entries[key] = value



from typing import Sequence

from kirin import ir, types, lowering
from kirin.decl import info, statement

from .runtime import IList
from ._dialect import dialect

ElemT = types.TypeVar("ElemT")
ListLen = types.TypeVar("ListLen")
IListType = types.Generic(IList, ElemT, ListLen)


@statement(dialect=dialect)
class Range(ir.Statement):
    name = "range"
    traits = frozenset({ir.Pure(), lowering.FromPythonRangeLike()})
    start: ir.SSAValue = info.argument(types.Int)
    stop: ir.SSAValue = info.argument(types.Int)
    step: ir.SSAValue = info.argument(types.Int)
    result: ir.ResultValue = info.result(IListType[types.Int, types.Any])


@statement(dialect=dialect, init=False)
class New(ir.Statement):
    traits = frozenset({ir.Pure(), lowering.FromPythonCall()})
    values: tuple[ir.SSAValue, ...] = info.argument(ElemT)
    elem_type: types.TypeAttribute = info.attribute()
    result: ir.ResultValue = info.result(IListType[ElemT])

    def __init__(
        self,
        values: Sequence[ir.SSAValue],
        elem_type: types.TypeAttribute | None = None,
    ) -> None:
        if not elem_type:
            elem_type = types.Any
        result_type = IListType[elem_type, types.Literal(len(values))]
        super().__init__(
            args=values,
            result_types=(result_type,),
            args_slice={"values": slice(0, len(values))},
            attributes={"elem_type": elem_type},
        )


@statement(dialect=dialect)
class Push(ir.Statement):
    traits = frozenset({ir.Pure(), lowering.FromPythonCall()})
    lst: ir.SSAValue = info.argument(IListType[ElemT])
    value: ir.SSAValue = info.argument(IListType[ElemT])
    result: ir.ResultValue = info.result(IListType[ElemT])


OutElemT = types.TypeVar("OutElemT")


@statement(dialect=dialect)
class Map(ir.Statement):
    traits = frozenset({ir.MaybePure(), lowering.FromPythonCall()})
    purity: bool = info.attribute(default=False)
    fn: ir.SSAValue = info.argument(types.MethodType[[ElemT], OutElemT])
    collection: ir.SSAValue = info.argument(IListType[ElemT, ListLen])
    result: ir.ResultValue = info.result(IListType[OutElemT, ListLen])


@statement(dialect=dialect)
class Foldr(ir.Statement):
    traits = frozenset({ir.MaybePure(), lowering.FromPythonCall()})
    purity: bool = info.attribute(default=False)
    fn: ir.SSAValue = info.argument(
        types.Generic(ir.Method, [ElemT, OutElemT], OutElemT)
    )
    collection: ir.SSAValue = info.argument(IListType[ElemT])
    init: ir.SSAValue = info.argument(OutElemT)
    result: ir.ResultValue = info.result(OutElemT)


@statement(dialect=dialect)
class Foldl(ir.Statement):
    traits = frozenset({ir.MaybePure(), lowering.FromPythonCall()})
    purity: bool = info.attribute(default=False)
    fn: ir.SSAValue = info.argument(
        types.Generic(ir.Method, [OutElemT, ElemT], OutElemT)
    )
    collection: ir.SSAValue = info.argument(IListType[ElemT])
    init: ir.SSAValue = info.argument(OutElemT)
    result: ir.ResultValue = info.result(OutElemT)


CarryT = types.TypeVar("CarryT")
ResultT = types.TypeVar("ResultT")


@statement(dialect=dialect)
class Scan(ir.Statement):
    traits = frozenset({ir.MaybePure(), lowering.FromPythonCall()})
    purity: bool = info.attribute(default=False)
    fn: ir.SSAValue = info.argument(
        types.Generic(ir.Method, [OutElemT, ElemT], types.Tuple[OutElemT, ResultT])
    )
    collection: ir.SSAValue = info.argument(IListType[ElemT, ListLen])
    init: ir.SSAValue = info.argument(OutElemT)
    result: ir.ResultValue = info.result(
        types.Tuple[OutElemT, IListType[ResultT, ListLen]]
    )


@statement(dialect=dialect)
class ForEach(ir.Statement):
    traits = frozenset({ir.MaybePure(), lowering.FromPythonCall()})
    purity: bool = info.attribute(default=False)
    fn: ir.SSAValue = info.argument(types.Generic(ir.Method, [ElemT], types.NoneType))
    collection: ir.SSAValue = info.argument(IListType[ElemT])



from kirin import ir, interp
from kirin.analysis import const
from kirin.dialects import func

from .stmts import Yield, IfElse
from ._dialect import dialect


@dialect.register(key="absint")
class Methods(interp.MethodTable):

    @interp.impl(Yield)
    def yield_stmt(
        self,
        interp_: interp.AbstractInterpreter,
        frame: interp.AbstractFrame,
        stmt: Yield,
    ):
        return interp.YieldValue(frame.get_values(stmt.values))

    @interp.impl(IfElse)
    def if_else(
        self,
        interp_: interp.AbstractInterpreter,
        frame: interp.AbstractFrame,
        stmt: IfElse,
    ):
        if isinstance(hint := stmt.cond.hints.get("const"), const.Value):
            if hint.data:
                return self._infer_if_else_cond(interp_, frame, stmt, stmt.then_body)
            else:
                return self._infer_if_else_cond(interp_, frame, stmt, stmt.else_body)
        then_results = self._infer_if_else_cond(interp_, frame, stmt, stmt.then_body)
        else_results = self._infer_if_else_cond(interp_, frame, stmt, stmt.else_body)

        match (then_results, else_results):
            case (interp.ReturnValue(then_value), interp.ReturnValue(else_value)):
                return interp.ReturnValue(then_value.join(else_value))
            case (interp.ReturnValue(then_value), _):
                return then_results
            case (_, interp.ReturnValue(else_value)):
                return else_results
            case _:
                return interp_.join_results(then_results, else_results)

    def _infer_if_else_cond(
        self,
        interp_: interp.AbstractInterpreter,
        frame: interp.AbstractFrame,
        stmt: IfElse,
        body: ir.Region,
    ):
        body_block = body.blocks[0]
        body_term = body_block.last_stmt
        if isinstance(body_term, func.Return):
            frame.worklist.append(interp.Successor(body_block, frame.get(stmt.cond)))
            return

        with interp_.new_frame(stmt, has_parent_access=True) as body_frame:
            ret = interp_.run_ssacfg_region(body_frame, body, (frame.get(stmt.cond),))
            frame.entries.update(body_frame.entries)
        return ret



from abc import abstractmethod
from types import GenericAlias
from typing import Any, Callable, Optional
from functools import cached_property
from dataclasses import MISSING, field, dataclass

from kirin import types
from kirin.ir import Block, Region, Attribute


@dataclass
class Field:
    name: Optional[str] = field(default=None, init=False)
    annotation: Any = field(default=None, init=False)
    kw_only: bool
    alias: Optional[str]

    __class_getitem__ = classmethod(GenericAlias)

    @abstractmethod
    def has_no_default(self) -> bool: ...


@dataclass
class AttributeField(Field):
    default: Any
    init: bool
    repr: bool
    default_factory: Optional[Callable[[], Attribute]]
    type: types.TypeAttribute
    pytype: bool = False
    "if `True`, annotation is a python type hint instead of `TypeAttribute`"

    def has_no_default(self):
        return self.default is MISSING and self.default_factory is None


def attribute(
    type: types.TypeAttribute = types.Any,
    *,
    init: bool = True,
    repr: bool = True,
    default: Any = MISSING,
    default_factory: Optional[Callable[[], Any]] = None,
    kw_only: bool = True,
    alias: Optional[str] = None,
) -> Any:
    if kw_only is False:
        raise TypeError("attribute fields must be keyword-only")

    return AttributeField(
        type=type,
        init=init,
        repr=repr,
        default=default,
        default_factory=default_factory,
        kw_only=kw_only,
        alias=alias,
    )


@dataclass
class ArgumentField(Field):
    type: types.TypeAttribute
    """type of the argument, will be used in validation.
    """
    print: bool = True
    """if `True`, this argument name is printed in the signature.
    """
    group: bool = False  # NOTE: this cannot be set by user
    """if `True`, this argument is annotated with Tuple[SSAValue, ...]
    """

    def has_no_default(self):
        return True


# NOTE: argument must appear in init and repr
def argument(
    type: types.TypeAttribute = types.Any,
    *,
    print: bool = True,
    kw_only: bool = False,
    alias: Optional[str] = None,
) -> Any:
    """Field specifier for arguments.

    Args:
        type(TypeAttribute): type of the argument, will be used in validation.
        print(bool): if `True`, this argument name is printed in the signature.
        kw_only(bool): if `True`, this argument is keyword-only.
        alias(Optional[str]): an alias for the argument name in the `__init__` method.
    """
    return ArgumentField(
        type=type,
        print=print,
        kw_only=kw_only,
        alias=alias,
    )


@dataclass
class ResultField(Field):
    init: bool
    repr: bool
    type: types.TypeAttribute = field(default_factory=types.AnyType)

    def has_no_default(self):
        return True


def result(
    type: types.TypeAttribute = types.Any,
    *,
    # NOTE: init is false, use other hooks to set custom results
    # or just mutate the statement after creation
    init: bool = False,
    repr: bool = True,
    kw_only: bool = True,
    alias: Optional[str] = None,
) -> Any:
    """Field specifier for results.

    Args:
        type(TypeAttribute): type of the result.
        init(bool): if `True`, this result field is included in the `__init__` method.
        repr(bool): if `True`, this result field is included in the `__repr__` and pretty printing.
        kw_only(bool): if `True`, this result field is keyword-only.
        alias(Optional[str]): an alias for the result field name in the `__init__` method.
    """
    if kw_only is False:  # for linting
        raise TypeError("result fields must be keyword-only")

    if init is True:
        raise TypeError("result fields cannot appear in __init__")

    return ResultField(
        type=type,
        init=init,
        repr=repr,
        kw_only=kw_only,
        alias=alias,
    )


@dataclass
class RegionField(Field):
    init: bool
    repr: bool
    multi: bool
    default_factory: Callable[[], Region]

    def has_no_default(self):
        return False


def region(
    *,
    init: bool = True,  # so we can use the default_factory
    repr: bool = True,
    kw_only: bool = True,
    alias: Optional[str] = None,
    multi: bool = False,
    default_factory: Callable[[], Region] = Region,
) -> Any:
    """Field specifier for regions.

    Args:
        init(bool): if `True`, this region field is included in the `__init__` method.
        repr(bool): if `True`, this region field is included in the `__repr__` and pretty printing.
        kw_only(bool): if `True`, this region field is keyword-only.
        alias(Optional[str]): an alias for the region field name in the `__init__` method.
        multi(bool): if `True`, this region can contain multiple blocks.
        default_factory(Callable[[], Region]): a factory function to create a default region.
    """
    if kw_only is False:
        raise TypeError("region fields must be keyword-only")

    return RegionField(
        init=init,
        repr=repr,
        kw_only=kw_only,
        alias=alias,
        multi=multi,
        default_factory=default_factory,
    )


@dataclass
class BlockField(Field):
    init: bool
    repr: bool
    default_factory: Callable[[], Block]

    def has_no_default(self):
        return False


def block(
    *,
    init: bool = True,
    repr: bool = True,
    kw_only: bool = True,
    alias: Optional[str] = None,
    default_factory: Callable[[], Block] = Block,
) -> Any:
    """Field specifier for blocks.

    Args:
        init(bool): if `True`, this block field is included in the `__init__` method.
        repr(bool): if `True`, this block field is included in the `__repr__` and pretty printing.
        kw_only(bool): if `True`, this block field is keyword-only.
        alias(Optional[str]): an alias for the block field name in the `__init__` method.
        default_factory(Callable[[], Block]): a factory function to create a default block.
    """
    if kw_only is False:
        raise TypeError("block fields must be keyword-only")

    return BlockField(
        init=init,
        repr=repr,
        kw_only=kw_only,
        alias=alias,
        default_factory=default_factory,
    )


@dataclass
class StatementFields:
    std_args: dict[str, ArgumentField] = field(default_factory=dict)
    """standard arguments of the statement."""
    kw_args: dict[str, ArgumentField] = field(default_factory=dict)
    """keyword-only arguments of the statement."""
    results: dict[str, ResultField] = field(default_factory=dict)
    """results of the statement."""
    regions: dict[str, RegionField] = field(default_factory=dict)
    """regions of the statement."""
    blocks: dict[str, BlockField] = field(default_factory=dict)
    """blocks of the statement."""
    attributes: dict[str, AttributeField] = field(default_factory=dict)
    """attributes of the statement."""

    class Args:
        def __init__(self, fields: "StatementFields"):
            self.fields = fields

        def __len__(self):
            return len(self.fields.std_args) + len(self.fields.kw_args)

        def __getitem__(self, name):
            if (value := self.fields.std_args.get(name)) is not None:
                return value
            elif (value := self.fields.kw_args.get(name)) is not None:
                return value
            raise KeyError(name)

        def __setitem__(self, name: str, value: ArgumentField):
            if value.kw_only:
                self.fields.kw_args[name] = value
            else:
                self.fields.std_args[name] = value

        def __contains__(self, name):
            return name in self.fields.std_args or name in self.fields.kw_args

        def values(self):
            yield from self.fields.std_args.values()
            yield from self.fields.kw_args.values()

        def items(self):
            yield from self.fields.std_args.items()
            yield from self.fields.kw_args.items()

        def keys(self):
            yield from self.fields.std_args.keys()
            yield from self.fields.kw_args.keys()

    @property
    def args(self):
        """iterable of all argument fields."""
        return self.Args(self)

    @classmethod
    def from_fields(cls, fields: dict[str, Field]):
        ret = cls()
        for name, f in fields.items():
            ret[name] = f
        return ret

    def __contains__(self, name):
        return (
            name in self.args
            or name in self.results
            or name in self.regions
            or name in self.blocks
            or name in self.attributes
        )

    def __setitem__(self, name, value):
        if isinstance(value, ArgumentField):
            self.args[name] = value
        elif isinstance(value, ResultField):
            self.results[name] = value
        elif isinstance(value, RegionField):
            self.regions[name] = value
        elif isinstance(value, BlockField):
            self.blocks[name] = value
        elif isinstance(value, AttributeField):
            self.attributes[name] = value
        else:
            raise TypeError(f"unknown field type {value}")

    def __iter__(self):
        yield from self.args.values()
        yield from self.kw_args.values()
        yield from self.results.values()
        yield from self.regions.values()
        yield from self.blocks.values()
        yield from self.attributes.values()

    def __len__(self):
        return (
            len(self.args)
            + len(self.results)
            + len(self.regions)
            + len(self.blocks)
            + len(self.attributes)
        )

    @cached_property
    def attr_or_props(self):
        return set(self.attributes.keys())

    @cached_property
    def required_names(self):
        """set of all fields that do not have a default value."""
        return set(
            list(self.args.keys())
            + [name for name, f in self.attributes.items() if f.has_no_default()]
            + [name for name, f in self.blocks.items() if f.has_no_default()]
            + [name for name, f in self.regions.items() if f.has_no_default()]
        )

    @cached_property
    def group_arg_names(self):
        return set([name for name, f in self.args.items() if f.group])



"""This module contains some common eDSLs that can be used to build
more complex eDSLs. The eDSLs in this module are built on top of the
basic dialects provided by the `kirin.dialects` module.
"""

from typing_extensions import Doc, Annotated

from kirin.ir import Method, dialect_group
from kirin.passes import Default
from kirin.dialects import cf, scf, func, math, ilist, lowering
from kirin.dialects.py import (
    cmp,
    len,
    attr,
    base,
    list,
    binop,
    slice,
    tuple,
    unary,
    assign,
    boolop,
    builtin,
    constant,
    indexing,
    iterable,
    assertion,
)


@dialect_group(
    [
        base,
        binop,
        cmp,
        unary,
        assign,
        attr,
        boolop,
        builtin,
        constant,
        indexing,
        len,
        tuple,
        assertion,
        iterable,
    ]
)
def python_basic(self):
    """The basic Python dialect without list, range, and slice."""

    def run_pass(mt: Method) -> None:
        pass

    return run_pass


@dialect_group(
    python_basic.union(
        [
            list,
            slice,
            cf,
            func,
            lowering.cf,
            lowering.func,
            lowering.call,
            lowering.range.ilist,
            math,
        ]
    )
)
def python_no_opt(self):
    """The Python dialect without optimization passes."""

    def run_pass(mt: Method) -> None:
        pass

    return run_pass


@dialect_group(
    python_basic.union(
        [
            ilist,
            slice,
            cf,
            func,
            math,
            lowering.cf,
            lowering.func,
            lowering.call,
            lowering.range.ilist,
        ]
    )
)
def basic_no_opt(self):
    """The basic kernel without optimization passes. This is a builtin
    eDSL that includes the basic dialects that are commonly used in
    Python-like eDSLs.

    This eDSL includes the basic dialects without any optimization passes.
    Other eDSL can usually be built on top of this eDSL by utilizing the
    `basic_no_opt.add` method to add more dialects and optimization passes.

    Note that unlike Python, list in this eDSL is immutable, and the
    `append` method is not available. Use `+` operator to concatenate lists
    instead. Immutable list is easier to optimize and reason about.

    See also [`basic`][kirin.prelude.basic] for the basic kernel with optimization passes.
    See also [`ilist`][kirin.dialects.ilist] for the immutable list dialect.
    """
    ilist_desugar = ilist.IListDesugar(self)

    def run_pass(mt: Method) -> None:
        ilist_desugar.fixpoint(mt)

    return run_pass


@dialect_group(basic_no_opt)
def basic(self):
    """The basic kernel.

    This eDSL includes the basic dialects and the basic optimization passes.
    Other eDSL can usually be built on top of this eDSL by utilizing the
    `basic.add` method to add more dialects and optimization passes.

    See also [`basic_no_opt`][kirin.prelude.basic_no_opt] for the basic kernel without optimization passes.

    ## Example

    ```python
    from kirin.prelude import basic

    @basic(typeinfer=True)
    def main(x: int) -> int:
        return x + 1 + 1

    main.print() # main is a Method!
    ```
    """

    def run_pass(
        mt: Annotated[Method, Doc("The method to run pass on.")],
        *,
        verify: Annotated[
            bool, Doc("run `verify` before running passes, default is `True`")
        ] = True,
        typeinfer: Annotated[
            bool,
            Doc(
                "run type inference and apply the inferred type to IR, default `False`"
            ),
        ] = False,
        fold: Annotated[bool, Doc("run folding passes")] = True,
        aggressive: Annotated[
            bool, Doc("run aggressive folding passes if `fold=True`")
        ] = False,
        no_raise: Annotated[bool, Doc("do not raise exception during analysis")] = True,
    ) -> None:
        default_pass = Default(
            self,
            verify=verify,
            fold=fold,
            aggressive=aggressive,
            typeinfer=typeinfer,
            no_raise=no_raise,
        )
        default_pass.fixpoint(mt)

    return run_pass


@dialect_group(
    python_basic.union(
        [
            ilist,
            slice,
            scf,
            cf,
            func,
            math,
            lowering.func,
            lowering.call,
            lowering.range.ilist,
        ]
    )
)
def structural_no_opt(self):
    """Structural kernel without optimization passes."""

    def run_pass(method: Method) -> None:
        pass

    return run_pass


@dialect_group(
    python_basic.union(
        [
            ilist,
            slice,
            scf,
            cf,
            func,
            math,
            lowering.func,
            lowering.call,
            lowering.range.ilist,
        ]
    )
)
def structural(self):
    """Structural kernel with optimization passes."""

    def run_pass(
        mt: Annotated[Method, Doc("The method to run pass on.")],
        *,
        verify: Annotated[
            bool, Doc("run `verify` before running passes, default is `True`")
        ] = True,
        typeinfer: Annotated[
            bool,
            Doc(
                "run type inference and apply the inferred type to IR, default `False`"
            ),
        ] = False,
        fold: Annotated[bool, Doc("run folding passes")] = True,
        aggressive: Annotated[
            bool, Doc("run aggressive folding passes if `fold=True`")
        ] = False,
        no_raise: Annotated[bool, Doc("do not raise exception during analysis")] = True,
    ) -> None:
        default_pass = Default(
            self,
            verify=verify,
            fold=fold,
            aggressive=aggressive,
            typeinfer=typeinfer,
            no_raise=no_raise,
        )
        default_pass.fixpoint(mt)

    return run_pass



from kirin import interp
from kirin.emit.julia import EmitJulia, EmitStrFrame

from . import stmts
from ._dialect import dialect


@dialect.register(key="emit.julia")
class JuliaTable(interp.MethodTable):

    @interp.impl(stmts.Not)
    def emit_Not(self, emit: EmitJulia, frame: EmitStrFrame, stmt: stmts.Not):
        return (emit.write_assign(frame, stmt.result, f"!{frame.get(stmt.value)}"),)

    @interp.impl(stmts.USub)
    def emit_USub(self, emit: EmitJulia, frame: EmitStrFrame, stmt: stmts.USub):
        return (emit.write_assign(frame, stmt.result, f"-{frame.get(stmt.value)}"),)

    @interp.impl(stmts.UAdd)
    def emit_UAdd(self, emit: EmitJulia, frame: EmitStrFrame, stmt: stmts.UAdd):
        return (emit.write_assign(frame, stmt.result, f"+{frame.get(stmt.value)}"),)



from abc import ABC, abstractmethod
from typing import ClassVar
from dataclasses import field, dataclass

from kirin.ir import Method, DialectGroup
from kirin.rewrite.abc import RewriteResult


@dataclass
class Pass(ABC):
    """A pass is a transformation that is applied to a method. It wraps
    the analysis and rewrites needed to transform the method as an independent
    unit.

    Unlike LLVM/MLIR passes, a pass in Kirin does not apply to a module,
    this is because we focus on individual methods defined within
    python modules. This is a design choice to allow seamless integration
    within the Python interpreter.

    A Kirin compile unit is a `ir.Method` object, which is always equivalent
    to a LLVM/MLIR module if it were lowered to LLVM/MLIR just like other JIT
    compilers.
    """

    name: ClassVar[str]
    dialects: DialectGroup
    no_raise: bool = field(default=True, kw_only=True)

    def __call__(self, mt: Method) -> RewriteResult:
        result = self.unsafe_run(mt)
        mt.code.verify()
        return result

    def fixpoint(self, mt: Method, max_iter: int = 32) -> RewriteResult:
        result = RewriteResult()
        for _ in range(max_iter):
            result_ = self.unsafe_run(mt)
            result = result_.join(result)
            if not result.has_done_something:
                break
        mt.verify()
        return result

    @abstractmethod
    def unsafe_run(self, mt: Method) -> RewriteResult: ...



from dataclasses import dataclass

from kirin import ir
from kirin.rewrite.abc import RewriteRule, RewriteResult


@dataclass
class DeadCodeElimination(RewriteRule):

    def rewrite_Statement(self, node: ir.Statement) -> RewriteResult:
        if self.is_pure(node):
            for result in node._results:
                if result.uses:
                    return RewriteResult()

            node.delete()
            return RewriteResult(has_done_something=True)

        return RewriteResult()



"""Pretty printing utilities.

This module provides a pretty printing utility for the IR nodes and other
objects in the compiler.

The pretty printing utility is implemented using the visitor pattern. The
[`Printable`][kirin.print.Printable] class is the base class for all objects that can be pretty printed.

The [`Printer`][kirin.print.Printer] class is the visitor that traverses the object and prints the
object to a string. The [`Printer`][kirin.print.Printer] class provides methods for printing different
types of objects.
"""

from kirin.print.printer import Printer as Printer
from kirin.print.printable import Printable as Printable



"""Copied from dataclasses in Python 3.10.13."""

from types import FunctionType


def set_qualname(cls: type, value):
    # Ensure that the functions returned from _create_fn uses the proper
    # __qualname__ (the class they belong to).
    if isinstance(value, FunctionType):
        value.__qualname__ = f"{cls.__qualname__}.{value.__name__}"
    return value


def set_new_attribute(cls: type, name: str, value):
    # Never overwrites an existing attribute.  Returns True if the
    # attribute already exists.
    if name in cls.__dict__:
        return True
    set_qualname(cls, value)
    setattr(cls, name, value)
    return False



from functools import partial

import pytest

from kirin import lowering
from kirin.prelude import python_no_opt
from kirin.dialects import cf, func

lower = lowering.Python(python_no_opt)


def test_closure():
    def closure(x):
        def inner(y):
            return x + y

        return inner

    code = lower.python_function(closure)
    assert isinstance(code, func.Function)

    def will_error(x):
        @partial
        def inner(y):
            return x + y

        return inner

    with pytest.raises(lowering.BuildError):
        lower.python_function(will_error)


def test_closure_branch():
    def closure_branch(m: int, n: int):

        def inner(x: int, y: int):
            if x > y:
                return m, n

        return inner

    code = lower.python_function(closure_branch)
    code.print()
    assert isinstance(code, func.Function)
    first_block = code.body.blocks[0]
    lambda_stmt = first_block.first_stmt
    assert isinstance(lambda_stmt, func.Lambda)
    assert lambda_stmt.args[0] is first_block.args[1]
    first_lambda_block = lambda_stmt.body.blocks[0]
    assert len(first_lambda_block.args) == 3
    get_field_n = first_lambda_block.stmts.at(0)
    assert isinstance(get_field_n, func.GetField)
    assert get_field_n.field == 1
    get_field_m = first_lambda_block.stmts.at(1)
    assert isinstance(get_field_m, func.GetField)
    assert get_field_m.field == 0
    assert isinstance(first_lambda_block.last_stmt, cf.ConditionalBranch)
    assert first_lambda_block.last_stmt.then_successor is lambda_stmt.body.blocks[1]
    assert first_lambda_block.last_stmt.else_successor is lambda_stmt.body.blocks[2]
    assert isinstance(lambda_stmt.body.blocks[1].last_stmt, func.Return)



from numbers import Number

import numpy as np

from kirin import ir, types, lowering
from kirin.decl import info, statement

from .dialect import _dialect


@statement
class PauliOperator(ir.Statement):
    traits = frozenset({ir.Pure(), lowering.FromPythonCall()})
    pre_factor: Number = info.attribute(default=1)
    result: ir.ResultValue = info.result(types.PyClass(np.matrix))


@statement(dialect=_dialect)
class X(PauliOperator):
    pass


@statement(dialect=_dialect)
class Y(PauliOperator):
    pass


@statement(dialect=_dialect)
class Z(PauliOperator):
    pass


@statement(dialect=_dialect)
class Id(PauliOperator):
    pass



import ast

from kirin import ir, lowering
from kirin.dialects.py import boolop

from . import stmts
from ._dialect import dialect


@dialect.register
class PythonLowering(lowering.FromPythonAST):

    def lower_Compare(
        self, state: lowering.State, node: ast.Compare
    ) -> lowering.Result:
        # NOTE: a key difference here is we need to lower
        # the multi-argument comparison operators into binary operators
        # since low-level comparision operators are binary + we need a static
        # number of arguments in each instruction
        lhs = state.lower(node.left).expect_one()

        comparators = [
            state.lower(comparator).expect_one() for comparator in node.comparators
        ]

        cmp_results: list[ir.SSAValue] = []
        for op, rhs in zip(node.ops, comparators):
            if cls := getattr(stmts, op.__class__.__name__, None):
                stmt: stmts.Cmp = cls(lhs=lhs, rhs=rhs)
            else:
                raise lowering.BuildError(f"unsupported compare operator {op}")
            state.current_frame.push(stmt)
            cmp_results.append(stmt.result)
            lhs = rhs

        if len(cmp_results) == 1:
            return cmp_results[0]

        lhs = cmp_results[0]
        for rhs in cmp_results[1:]:
            and_stmt = boolop.And(lhs=lhs, rhs=rhs)
            state.current_frame.push(and_stmt)
            lhs = and_stmt.result

        return lhs



from kirin import types
from kirin.prelude import structural_no_opt
from kirin.analysis import TypeInference
from kirin.dialects import ilist

type_infer = TypeInference(structural_no_opt)


def test_inside_return_loop():
    @structural_no_opt
    def simple_loop(x: float):
        for i in range(0, 3):
            return i
        return x

    frame, ret = type_infer.run_analysis(simple_loop)
    assert ret.is_subseteq(types.Int | types.Float)


def test_simple_ifelse():
    @structural_no_opt
    def simple_ifelse(x: int):
        cond = x > 0
        if cond:
            return cond
        else:
            return 0

    frame, ret = type_infer.run_analysis(simple_ifelse)
    assert ret.is_subseteq(types.Bool | types.Int | types.NoneType)


def test_getitem_inside_loop():
    @structural_no_opt
    def getitem_loop():
        ls = [1, 2, 3]
        for i in range(2):
            ls[i + 1]

    frame, ret = type_infer.run_analysis(getitem_loop)

    getitem_loop.print(analysis=frame.entries)

    # NOTE: go through values, but skip method and return
    for value in list(frame.entries.values())[1:-1]:
        # everything except the method & return should either be int or a list[int]
        print(value)
        print(value.is_subseteq(types.Int | ilist.IListType[types.Int, types.Any]))

    assert ret.is_subseteq(types.NoneType)

    @structural_no_opt
    def getitem_return_from_loop():
        ls = [1, 2, 3]
        for i in range(2):
            return ls[i + 1]

        return 1.0

    frame, ret = type_infer.run_analysis(getitem_return_from_loop)
    assert ret.is_subseteq(types.Int | types.Float)



from kirin import types
from kirin.prelude import structural_no_opt
from kirin.analysis import TypeInference


def test_self_ref_closure():

    @structural_no_opt
    def should_work(n_qubits: int):
        def self_ref_source(i_layer):
            stride = n_qubits // (2**i_layer)
            if stride == 0:
                return

            self_ref_source(i_layer + 1)

        return self_ref_source

    infer = TypeInference(structural_no_opt)
    _, ret = infer.run_analysis(should_work, (types.Int,))
    assert ret.is_equal(types.MethodType[types.Tuple[types.Any], types.NoneType])



import pytest

from kirin import ir, lowering
from kirin.decl import statement
from kirin.prelude import basic_no_opt

dialect = ir.Dialect("foo")


@statement(dialect=dialect)
class InvalidStmt(ir.Statement):
    traits = frozenset({lowering.FromPythonCall()})

    def check(self):
        raise ValueError("Never triggers")


@statement(dialect=dialect)
class InvalidType(ir.Statement):
    traits = frozenset({lowering.FromPythonCall()})

    def check_type(self):
        raise ValueError("Never triggers")


@ir.dialect_group(basic_no_opt.add(dialect))
def foo(self):
    def run_pass(mt):
        pass

    return run_pass


def test_invalid_stmt():
    @foo
    def test():
        InvalidStmt()

    with pytest.raises(Exception):
        test.verify()


def test_invalid_type():
    @foo
    def test():
        InvalidType()

    with pytest.raises(Exception):
        test.verify_type()



from kirin.prelude import basic_no_opt


@basic_no_opt
def factorial(n):
    if n == 0:
        return 1
    else:
        return n * factorial(n - 1)


def test_region_clone():
    assert factorial.callable_region.clone().is_structurally_equal(
        factorial.callable_region
    )



from __future__ import annotations

from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Generic, TypeVar
from dataclasses import dataclass

if TYPE_CHECKING:
    from kirin.ir import Block, Region, Attribute, Statement
    from kirin.graph import Graph


IRNodeType = TypeVar("IRNodeType")


@dataclass(frozen=True)
class Trait(ABC, Generic[IRNodeType]):
    """Base class for all statement traits."""

    def verify(self, node: IRNodeType):
        pass


@dataclass(frozen=True)
class AttrTrait(Trait["Attribute"]):
    """Base class for all attribute traits."""

    def verify(self, node: "Attribute"):
        pass


@dataclass(frozen=True)
class StmtTrait(Trait["Statement"], ABC):
    """Base class for all statement traits."""

    def verify(self, node: "Statement"):
        pass


GraphType = TypeVar("GraphType", bound="Graph[Block]")


@dataclass(frozen=True)
class RegionTrait(StmtTrait, Generic[GraphType]):
    """A trait that indicates the properties of the statement's region."""

    @abstractmethod
    def get_graph(self, region: Region) -> GraphType: ...



import os
import logging
import argparse
import subprocess

import tomlkit
from packaging.version import Version

# check if repo is dirty
if subprocess.run(["git", "diff", "--quiet"]).returncode != 0:
    logging.error("Repo is dirty")
    exit(1)

parser = argparse.ArgumentParser(description="Release a new version of the package")
parser.add_argument("version", type=str, help="The version to release")
args = parser.parse_args()

root_dir = os.path.dirname(os.path.dirname(__file__))
toml_path = os.path.join(root_dir, "pyproject.toml")
dist_dir = os.path.join(root_dir, "dist")

with open(toml_path, "r") as f:
    data = tomlkit.parse(f.read())

v = Version(data["project"]["version"])

if args.version == "patch":
    new_v = f"{v.major}.{v.minor}.{v.micro + 1}"
elif args.version == "minor":
    new_v = f"{v.major}.{v.minor + 1}.0"
elif args.version == "major":
    new_v = f"{v.major + 1}.0.0"
elif args.version == "current":
    new_v = str(v)

# query if the new version exists
print(f"gh release view {new_v}")
if subprocess.run(["gh", "release", "view", new_v]).returncode == 0:
    logging.error(f"Version {new_v} already exists")
    exit(1)

modified = new_v != str(v)
data["project"]["version"] = new_v
with open(toml_path, "w") as f:
    tomlkit.dump(data, f)


def fail_and_reset(msg):
    logging.error(msg)
    data["project"]["version"] = str(v)
    with open(toml_path, "w") as f:
        tomlkit.dump(data, f)
    exit(1)


# 0. commit version bump
# directly output git error msg

commited = False
if modified:
    print("git add project.toml")
    if subprocess.run(["git", "add", "pyproject.toml"]).returncode != 0:
        fail_and_reset("Failed to add pyproject.toml")

if modified:
    if subprocess.run(["git", "add", "pyproject.toml"]).returncode != 0:
        fail_and_reset("Failed to add pyproject.toml")

    print(f"git commit -m 'Bump version to {new_v}'")
    if (
        subprocess.run(["git", "commit", "-m", f"Bump version to {new_v}"]).returncode
        != 0
    ):
        fail_and_reset("Failed to commit version bump")


def fail_and_revert(msg):
    logging.error(msg)
    if commited:
        print("Reverting commit")
        subprocess.run(["git", "reset", "--hard", "HEAD~1"])
    exit(1)


try:
    # sync with remote no matter what
    if subprocess.run(["git", "push"]).returncode != 0:
        fail_and_revert("Failed to push")

    # 1. build the package via uv
    print("uv build")
    if subprocess.run(["uv", "build"]).returncode != 0:
        fail_and_revert("Failed to build the package")

    # 2. create a new release
    print(f"gh release create {new_v} dist/*")
    if (
        subprocess.run(
            [
                "gh",
                "release",
                "create",
                "v" + new_v,
            ]
        ).returncode
        != 0
    ):
        fail_and_revert("Failed to create release")

except Exception as e:
    fail_and_revert(str(e))



from dataclasses import field, dataclass

from kirin import ir, passes, rewrite
from kirin.analysis import CallGraph
from kirin.rewrite.abc import RewriteRule, RewriteResult
from kirin.dialects.func.stmts import Invoke


@dataclass
class ReplaceMethods(RewriteRule):
    new_symbols: dict[ir.Method, ir.Method]

    def rewrite_Statement(self, node: ir.Statement) -> RewriteResult:
        if (
            not isinstance(node, Invoke)
            or (new_callee := self.new_symbols.get(node.callee)) is None
        ):
            return RewriteResult()

        node.replace_by(
            Invoke(
                inputs=node.inputs,
                callee=new_callee,
                purity=node.purity,
                kwargs=node.kwargs,
            )
        )

        return RewriteResult(has_done_something=True)


@dataclass
class CallGraphPass(passes.Pass):
    """Copy all functions in the call graph and apply a rule to each of them.


    Usage:
        rule = Walk(SomeRewriteRule())
        pass_ = CallGraphPass(rule=rule, dialects=...)
        pass_(some_method)

    Note: This pass modifies the input method in place, but copies
    all methods invoked within it before applying the rule to them.

    """

    rule: RewriteRule
    """The rule to apply to each function in the call graph."""

    fold_pass: passes.Fold = field(init=False)

    def __post_init__(self):
        self.fold_pass = passes.Fold(self.dialects, no_raise=self.no_raise)

    def unsafe_run(self, mt: ir.Method) -> RewriteResult:
        result = RewriteResult()
        mt_map = {}

        cg = CallGraph(mt)

        all_methods = set(sum(map(tuple, cg.defs.values()), ()))
        for original_mt in all_methods:
            if original_mt is mt:
                new_mt = original_mt
            else:
                new_mt = original_mt.similar()
            result = self.rule.rewrite(new_mt.code).join(result)
            mt_map[original_mt] = new_mt

        if result.has_done_something:
            for _, new_mt in mt_map.items():
                rewrite.Walk(ReplaceMethods(mt_map)).rewrite(new_mt.code)
                self.fold_pass(new_mt)

        return result



from kirin import ir
from kirin.analysis import const
from kirin.dialects import func
from kirin.rewrite.abc import RewriteRule, RewriteResult
from kirin.dialects.py.constant import Constant

from .stmts import For, Yield, IfElse


class PickIfElse(RewriteRule):

    def rewrite_Statement(self, node: ir.Statement) -> RewriteResult:
        if not isinstance(node, IfElse):
            return RewriteResult()

        if not isinstance(hint := node.cond.hints.get("const"), const.Value):
            return RewriteResult()

        if hint.data:
            return self.insert_body(node, node.then_body)
        else:
            return self.insert_body(node, node.else_body)

    def insert_body(self, node: IfElse, body: ir.Region):
        body_block = body.blocks[0]
        body_block.args[0].replace_by(node.cond)
        block_stmt = body_block.first_stmt
        while block_stmt and not block_stmt.has_trait(ir.IsTerminator):
            block_stmt.detach()
            block_stmt.insert_before(node)
            block_stmt = body_block.first_stmt

        terminator = body_block.last_stmt
        if isinstance(terminator, Yield):
            for result, output in zip(node.results, terminator.values):
                result.replace_by(output)
            node.delete()
            return RewriteResult(has_done_something=True)
        elif isinstance(terminator, func.Return):
            block = node.parent
            assert block is not None
            stmt = block.last_stmt
            while stmt is not None and stmt is not node:  # remove the rest of the block
                delete_stmt = stmt
                stmt = stmt.prev_stmt
                delete_stmt.delete()

            terminator.detach()
            node.replace_by(terminator)
            return RewriteResult(has_done_something=True)
        else:
            return RewriteResult()


class ForLoop(RewriteRule):

    def rewrite_Statement(self, node: ir.Statement) -> RewriteResult:
        if not isinstance(node, For):
            return RewriteResult()

        # TODO: support for PartialTuple and IList with known length
        if not isinstance(hint := node.iterable.hints.get("const"), const.Value):
            return RewriteResult()

        loop_vars = node.initializers
        for item in hint.data:
            body = node.body.clone()
            block = body.blocks[0]
            item_stmt = Constant(item)
            item_stmt.insert_before(node)
            block.args[0].replace_by(item_stmt.result)
            for var, input in zip(block.args[1:], loop_vars):
                var.replace_by(input)

            block_stmt = block.first_stmt
            while block_stmt and not block_stmt.has_trait(ir.IsTerminator):
                block_stmt.detach()
                block_stmt.insert_before(node)
                block_stmt = block.first_stmt

            terminator = block.last_stmt
            # we assume Yield has the same # of values as initializers
            # TODO: check this in validation
            if isinstance(terminator, Yield):
                loop_vars = terminator.values
                terminator.delete()

        for result, output in zip(node.results, loop_vars):
            result.replace_by(output)
        node.delete()
        return RewriteResult(has_done_something=True, terminated=True)



import pytest

from kirin import ir, types, rewrite
from kirin.prelude import python_basic
from kirin.dialects import py, scf, func, ilist, lowering


def test_cons():
    x0 = py.Constant(0)
    iter = py.Constant(range(5))
    body = ir.Region(ir.Block([]))
    idx = body.blocks[0].args.append_from(types.Any, "idx")
    body.blocks[0].args.append_from(types.Any, "acc")
    body.blocks[0].stmts.append(scf.Yield(idx))
    stmt = scf.For(iter.result, body, x0.result)
    assert len(stmt.results) == 1

    body = ir.Region(ir.Block([]))
    idx = body.blocks[0].args.append_from(types.Any, "idx")
    body.blocks[0].stmts.append(scf.Yield(idx))

    with pytest.raises(ir.ValidationError):
        stmt = scf.For(iter.result, body, x0.result)
        stmt.verify()

    body = ir.Region(ir.Block([]))
    idx = body.blocks[0].args.append_from(types.Any, "idx")
    with pytest.raises(ir.ValidationError):
        stmt = scf.For(iter.result, body, x0.result)
        stmt.verify()


def test_exec():
    xs = ilist.IList([(1, 2), (3, 4)])

    @python_basic.union(
        [func, scf, py.unpack, ilist, lowering.func, lowering.range.ilist]
    )
    def main(x):
        for a, b in xs:
            x = x + a
        return x

    main.print()
    assert main(0) == 4


def test_issue_213():

    @python_basic.union(
        [func, scf, py.unpack, ilist, lowering.func, lowering.range.ilist]
    )
    def main():
        j = 0.0
        i = 0
        for k in range(2):
            j = j + i + k

        for k in range(2):
            j = j + i

        return j

    assert main.py_func is not None
    assert main() == main.py_func()


def test_simple_assign():
    @python_basic.union(
        [func, scf, py.unpack, lowering.func, ilist, lowering.range.ilist]
    )
    def main(n: int):
        x = 0
        for i in range(n):
            x = i
        return x

    assert main(5) == 4


def test_unused_loop_vars():
    @python_basic.union(
        [func, scf, py.unpack, lowering.func, ilist, lowering.range.ilist]
    )
    def main(n: int):
        x = 0
        offset = 1
        for i in range(n):
            x = i + offset
        return x

    rewrite.Walk(scf.trim.UnusedYield()).rewrite(main.code)
    loop = main.callable_region.blocks[0].stmts.at(-2)
    assert isinstance(loop, scf.For)
    assert len(loop.initializers) == 1
    assert len(loop.body.blocks[0].args) == 2
    assert main(5) == 5


def test_unused_loop_vars_adding_ints():
    @python_basic.union(
        [func, scf, py.unpack, lowering.func, ilist, lowering.range.ilist]
    )
    def main(n: int):
        x = 0
        for i in range(n):
            x += i
        return x

    rule = scf.trim.UnusedYield()
    rewrite.Walk(rule).rewrite(main.code)
    loop = main.callable_region.blocks[0].stmts.at(-2)
    assert isinstance(loop, scf.For)
    assert len(loop.initializers) == 1
    assert len(loop.body.blocks[0].args) == 2
    assert main(4) == 6


def test_unused_loop_vars_sum_ilist():
    @python_basic.union(
        [func, scf, py.unpack, lowering.func, ilist, lowering.range.ilist]
    )
    def main():
        data = [1, 2, 3]
        total = 0
        for value in data:
            total += value
        return total

    rule = scf.trim.UnusedYield()
    rewrite.Walk(rule).rewrite(main.code)
    loop = main.callable_region.blocks[0].stmts.at(-2)
    assert isinstance(loop, scf.For)
    assert len(loop.initializers) == 1
    assert len(loop.body.blocks[0].args) == 2
    assert main() == 6


def test_unused_loop_vars_multiple_mutations():
    @python_basic.union(
        [func, scf, py.unpack, lowering.func, ilist, lowering.range.ilist]
    )
    def main(n: int):
        """Sum integers from 0 to n - 1"""
        total = 0
        total_2 = 0

        for value in range(n):
            total += value
            total_2 += 2 * total
        return total, total_2

    rule = scf.trim.UnusedYield()
    rewrite.Walk(rule).rewrite(main.code)

    loop = main.callable_region.blocks[0].stmts.at(-3)
    assert isinstance(loop, scf.For)
    assert len(loop.initializers) == 2
    assert len(loop.body.blocks[0].args) == 3
    assert main(4) == (6, 20)


def test_body_with_no_yield():

    @python_basic.union(
        [func, scf, py.unpack, lowering.func, ilist, lowering.range.ilist]
    )
    def julia_like(x: int, y: int):
        for i in range(x):
            i == 0
        return x + y

    for_stmt = next(
        stmt for stmt in julia_like.callable_region.stmts() if isinstance(stmt, scf.For)
    )
    for_body_stmts = list(for_stmt.body.stmts())
    assert not isinstance(for_body_stmts[-1], scf.Yield)



from kirin import ir
from kirin.decl import statement
from kirin.prelude import structural_no_opt
from kirin.rewrite import abc
from kirin.dialects import func
from kirin.passes.callgraph import CallGraphPass
from kirin.analysis.callgraph import CallGraph


def test_callgraph_pass():

    dialect = ir.Dialect("test")

    @statement(dialect=dialect)
    class TestStmt(ir.Statement):
        pass

    class RewriteRule(abc.RewriteRule):
        def rewrite_Statement(self, node: ir.Statement) -> abc.RewriteResult:
            if not isinstance(node, func.Function):
                return abc.RewriteResult()

            first_stmt = node.body.blocks[0].first_stmt
            assert first_stmt is not None
            TestStmt().insert_before(first_stmt)

            return abc.RewriteResult(has_done_something=True)

    @structural_no_opt
    def subroutine(x):
        if x < 10:
            return subroutine(x + 1)
        else:
            return x

    @structural_no_opt
    def main():
        return subroutine(0)

    rule = RewriteRule()
    pass_ = CallGraphPass(rule=rule, dialects=structural_no_opt)

    pass_(main)

    cg = CallGraph(main)

    for mts in cg.defs.values():
        for mt in mts:
            assert isinstance(mt.callable_region.blocks[0].stmts.at(0), TestStmt)



from abc import ABC
from typing import IO, Generic, TypeVar
from dataclasses import field, dataclass

from kirin import ir, interp, idtable
from kirin.emit.abc import EmitABC, EmitFrame

IO_t = TypeVar("IO_t", bound=IO)


@dataclass
class EmitStrFrame(EmitFrame[str]):
    indent: int = 0
    captured: dict[ir.SSAValue, tuple[str, ...]] = field(default_factory=dict)


@dataclass
class EmitStr(EmitABC[EmitStrFrame, str], ABC, Generic[IO_t]):
    void = ""
    file: IO_t
    prefix: str = field(default="", kw_only=True)
    prefix_if_none: str = field(default="var_", kw_only=True)

    def initialize(self):
        super().initialize()
        self.ssa_id = idtable.IdTable[ir.SSAValue](
            prefix=self.prefix, prefix_if_none=self.prefix_if_none
        )
        self.block_id = idtable.IdTable[ir.Block](prefix=self.prefix + "block_")
        return self

    def initialize_frame(
        self, code: ir.Statement, *, has_parent_access: bool = False
    ) -> EmitStrFrame:
        return EmitStrFrame(code, has_parent_access=has_parent_access)

    def run_method(
        self, method: ir.Method, args: tuple[str, ...]
    ) -> tuple[EmitStrFrame, str]:
        if self.state.depth >= self.max_depth:
            raise interp.InterpreterError("maximum recursion depth exceeded")
        return self.run_callable(method.code, (method.sym_name,) + args)

    def write(self, *args):
        for arg in args:
            self.file.write(arg)

    def newline(self, frame: EmitStrFrame):
        self.file.write("\n" + "  " * frame.indent)

    def writeln(self, frame: EmitStrFrame, *args):
        self.newline(frame)
        self.write(*args)
